/****************************************************************************
 *   COPYRIGHT 2019 All rights reserved
 *   Panasonic Industrial Devices Europe GmbH
 ****************************************************************************/

/* This is *not* a generic compatiblity layer - it is specific to the demo
application. For example, some settings at some places are hardcoded, like PWM3.
*/

#include "compat.h"

static void (*gpio_interrupt_function)(uint32_t);

static void gpio_interrupt(void)
{
  gpio_interrupt_function(0);
}

static int32_t GPIO_Initialize(GPIO_Port_t port, GPIO_Pin_Configuration_t *configuration, void (*func)(uint32_t))
{
  RET_API ret;

  if (!func) {
    ret = gpio_config(GPIO10, configuration);
    if (ret) {
      return -1;
    }

    return 0;
  }

  /* fixme: only one interupt supported */
  ret = gpio_external_interrupt_init(port, configuration, gpio_interrupt);
  if (ret) {
    return -1;
  }

  gpio_interrupt_function = func;

  return 0;
}

static int32_t GPIO_Write(GPIO_Port_t port, GPIO_Bit_Value_t value)
{
  gpio_write(port, value);
  return 0;
}

TC35680_DRIVER_GPIO Driver_GPIO = {GPIO_Initialize, GPIO_Write};

static void (*PWM_CB_Function)(uint32_t);

/* taken from internal.h of the tc35678 SDK directory */
#define REG_RD(adr) (*(volatile unsigned int *)((unsigned long)(adr)))
#define REG_WR32(x, y) (*(volatile unsigned int *)((unsigned int *)(x))) = ((unsigned int)(y))
#define PSSEXT_ADR 0x4004A100L
#define PWM_INTCLR_CH3 (PSSEXT_ADR + 0x318)
#define PWM_PCTRL_CH3 (PSSEXT_ADR + 0x300)

static void pwm_interrupt(void)
{
  /* fixme: channel 3 hardcoded */
  REG_WR32(PWM_INTCLR_CH3, 1);
  PWM_CB_Function(0);
}

RET_API PWM_Initialize(PWM_Channel_t channel, PWM_Clock_t clock, PWM_Polarity_t polarity, uint16_t DivNum_16mHz, void (*CB_Function)(uint32_t))
{
  RET_API ret;

  PWM_CB_Function = CB_Function;

  ret = pwm_init(channel, (enum pwm_clock_select)clock);
  if (ret != RET_API_NORMAL) {
    return ret;
  }

  /* pwm driver does not enable interrupts, so we do it here */
  REG_WR32(PWM_PCTRL_CH3, (REG_RD(PWM_PCTRL_CH3) & ~(1 << 3)));

  /* fixme: channel 3 hardcoded */
  ret = SYS_API_SetCallbackPeripheralInt(pwm_interrupt, 0x400);
  if (ret != RET_API_NORMAL) {
    return ret;
  }

  return RET_API_NORMAL;
}

RET_API PWM_Config(PWM_Channel_t channel, uint16_t cycle, uint16_t duty, uint32_t rhythm_pattern)
{
  struct pwm_config_struct config_struct;
  RET_API ret;

  config_struct.rhythm_pattern = rhythm_pattern;
  config_struct.cycle = cycle;
  config_struct.duty = duty;

  ret = pwm_config(channel, &config_struct);
  if (ret != RET_API_NORMAL) {
    return ret;
  }

  return RET_API_NORMAL;
}

#define PWM_SetRhythmCounter pwm_set_rhythm_counter
#define PWM_StartPWM pwm_start
#define PWM_StopPWM pwm_stop

TC35680_DRIVER_PWM Driver_PWM = {PWM_Initialize, PWM_Config, PWM_SetRhythmCounter, PWM_StartPWM, PWM_StopPWM};

#include "ble_api.h"

enum ble_api_result ble_api_mng_config_connection_count(uint8_t remote_master_count, uint8_t remote_slave_count)
{
  uint8_t status = 0;

  ble_application_mng_config_connection_count_response_callback(status);

  return LE_API_SUCCESS;
}
