/****************************************************************************
 *   COPYRIGHT 2019 All rights reserved
 *   Panasonic Industrial Devices Europe GmbH
 ****************************************************************************/

//#define DEBUG

#include "application.h"
#include "debug.h"

#include "music.h"

enum music_tunes music_tune;

/* tune handling */

/* MIDI octave 5 frequencies */
static const uint16_t notes[] = {
    5233, 5544, 5873, 6223, 6593, 6985, 7400, 7840, 8306, 8880, 9323, 9878,
};

/* readable names for the musical scale */
enum {
  NOTE_C = 0,
  NOTE_Cs = 1,
  NOTE_D = 2,
  NOTE_Ds = 3,
  NOTE_E = 4,
  NOTE_F = 5,
  NOTE_Fs = 6,
  NOTE_G = 7,
  NOTE_Gs = 8,
  NOTE_A = 9,
  NOTE_As = 10,
  NOTE_B = 11,
  PAUSE = 12,
};

/* a single tone - note + octave + duration */
struct tone {
  uint8_t note;
  int8_t octave;
  uint16_t duration;
};

#define MS_OF_QUARTER_NOTE_FROM_BPM(x) ((60 * 1000) / x / 4)

#define DUR_SIXTEENTH 1
#define DUR_EIGTH 2
#define DUR_QUARTER 4
#define DUR_QUARTER_DOT 6
#define DUR_HALF 8
#define DUR_HALF_DOT 12
#define DUR_FULL 16

struct tune {
  int bpm;
  struct tone *tones;
  int length;
};

// clang-format off
struct tone tones_ode_an_die_freude[] = 
{
  {NOTE_E,  4, DUR_QUARTER, },
  {NOTE_E,  4, DUR_QUARTER, },
  {NOTE_F,  4, DUR_QUARTER, },
  {NOTE_G,  4, DUR_QUARTER, },
        
  {NOTE_G,  4, DUR_QUARTER, },
  {NOTE_F,  4, DUR_QUARTER, },
  {NOTE_E,  4, DUR_QUARTER, },
  {NOTE_D,  4, DUR_QUARTER, },
        
  {NOTE_C,  4, DUR_QUARTER, },
  {NOTE_C,  4, DUR_QUARTER, },
  {NOTE_D,  4, DUR_QUARTER, },
  {NOTE_E,  4, DUR_QUARTER, },

  {NOTE_E,  4, DUR_QUARTER_DOT, },
  {NOTE_D,  4, DUR_EIGTH, },
  {NOTE_D,  4, DUR_HALF, },

  {NOTE_E,  4, DUR_QUARTER, },
  {NOTE_E,  4, DUR_QUARTER, },
  {NOTE_F,  4, DUR_QUARTER, },
  {NOTE_G,  4, DUR_QUARTER, },
        
  {NOTE_G,  4, DUR_QUARTER, },
  {NOTE_F,  4, DUR_QUARTER, },
  {NOTE_E,  4, DUR_QUARTER, },
  {NOTE_D,  4, DUR_QUARTER, },

  {NOTE_C,  4, DUR_QUARTER, },
  {NOTE_C,  4, DUR_QUARTER, },
  {NOTE_D,  4, DUR_QUARTER, },
  {NOTE_E,  4, DUR_QUARTER, },

  {NOTE_D,  4, DUR_QUARTER_DOT, },
  {NOTE_C,  4, DUR_EIGTH, },
  {NOTE_C,  4, DUR_HALF, },
};

struct tune tune_ode_an_die_freude = 
{
  200, 
  &tones_ode_an_die_freude[0], 
  sizeof(tones_ode_an_die_freude) / sizeof(tones_ode_an_die_freude[0])
};

struct tone tones_fuer_elise[] = 
{
  {NOTE_E,  5, DUR_EIGTH,},
  {NOTE_Ds, 5, DUR_EIGTH,},

  {NOTE_E,  5, DUR_EIGTH,},
  {NOTE_Ds, 5, DUR_EIGTH,},
  {NOTE_E,  5, DUR_EIGTH,},
  {NOTE_B,  4, DUR_EIGTH,},
  {NOTE_D,  5, DUR_EIGTH,},
  {NOTE_C,  5, DUR_EIGTH,},

  {NOTE_A,  4, DUR_EIGTH,},
  {NOTE_E,  3, DUR_EIGTH,},
  {NOTE_A,  3, DUR_EIGTH,},
  {NOTE_C,  4, DUR_EIGTH,},
  {NOTE_E,  4, DUR_EIGTH,},
  {NOTE_A,  4, DUR_EIGTH,},

  {NOTE_B,  4, DUR_EIGTH,},
  {NOTE_E,  3, DUR_EIGTH,},
  {NOTE_Gs, 3, DUR_EIGTH,},
  {NOTE_E,  4, DUR_EIGTH,},
  {NOTE_Gs, 4, DUR_EIGTH,},
  {NOTE_B,  4, DUR_EIGTH,},

  {NOTE_C,  5, DUR_EIGTH,},
  {NOTE_E,  3, DUR_EIGTH,},
  {NOTE_A,  3, DUR_EIGTH,},
  {NOTE_E,  4, DUR_EIGTH,},
  {NOTE_E,  5, DUR_EIGTH,},
  {NOTE_Ds, 5, DUR_EIGTH,},

  {NOTE_E,  5, DUR_EIGTH,},
  {NOTE_Ds, 5, DUR_EIGTH,},
  {NOTE_E,  5, DUR_EIGTH,},
  {NOTE_B,  4, DUR_EIGTH,},
  {NOTE_D,  5, DUR_EIGTH,},
  {NOTE_C,  5, DUR_EIGTH,},
        
  {NOTE_A,  4, DUR_EIGTH,},
  {NOTE_E,  3, DUR_EIGTH,},
  {NOTE_A,  3, DUR_EIGTH,},
  {NOTE_C,  4, DUR_EIGTH,},
  {NOTE_E,  4, DUR_EIGTH,},
  {NOTE_A,  4, DUR_EIGTH,},      
        
  {NOTE_B,  4, DUR_EIGTH,},
  {NOTE_E,  3, DUR_EIGTH,},
  {NOTE_Gs, 3, DUR_EIGTH,},
  {NOTE_E,  4, DUR_EIGTH,},
  {NOTE_C,  5, DUR_EIGTH,},
  {NOTE_B,  4, DUR_EIGTH,},        

  {NOTE_A,  4, DUR_EIGTH,},
  {NOTE_E,  3, DUR_EIGTH,},
  {NOTE_A,  3, DUR_EIGTH,},
}; 

struct tune tune_fuer_elise = 
{
  150,
  &tones_fuer_elise[0],
  sizeof(tones_fuer_elise) / sizeof(tones_fuer_elise[0])
};

struct tone tones_kleine_nachtmusik[] = {
  {NOTE_G,  4, DUR_QUARTER_DOT,},
  {NOTE_D,  4, DUR_EIGTH,},
  {NOTE_G,  4, DUR_QUARTER_DOT,},
  {NOTE_D,  4, DUR_EIGTH,},

  {NOTE_G,  4, DUR_EIGTH,},
  {NOTE_D,  4, DUR_EIGTH,},
  {NOTE_G,  4, DUR_EIGTH,},
  {NOTE_B,  4, DUR_EIGTH,},
  {NOTE_D,  5, DUR_HALF,},

  {NOTE_C,  5, DUR_QUARTER_DOT,},
  {NOTE_A,  4, DUR_EIGTH,},
  {NOTE_C,  5, DUR_QUARTER_DOT,},
  {NOTE_A,  4, DUR_EIGTH,},

  {NOTE_C,  5, DUR_EIGTH,},
  {NOTE_A,  4, DUR_EIGTH,},
  {NOTE_Fs, 4, DUR_EIGTH,},
  {NOTE_A,  4, DUR_EIGTH,},
  {NOTE_D,  4, DUR_HALF,},

  {NOTE_G,  4, DUR_QUARTER,},
  {NOTE_G,  4, DUR_HALF,},
  {NOTE_A,  4, DUR_EIGTH,},
  {NOTE_G,  4, DUR_EIGTH,},

  {NOTE_G,  4, DUR_EIGTH,},
  {NOTE_Fs, 4, DUR_EIGTH,},
  {NOTE_Fs, 4, DUR_HALF,},
  {NOTE_C,  5, DUR_EIGTH,},
  {NOTE_Fs, 4, DUR_EIGTH,},

  {NOTE_A,  4, DUR_EIGTH,},
  {NOTE_G,  4, DUR_EIGTH,},
  {NOTE_G,  4, DUR_HALF,},
  {NOTE_A,  4, DUR_EIGTH,},
  {NOTE_G,  4, DUR_EIGTH,},

  {NOTE_G,  4, DUR_EIGTH,},
  {NOTE_Fs, 4, DUR_EIGTH,},
  {NOTE_Fs, 4, DUR_HALF,},
  {NOTE_C,  5, DUR_EIGTH,},
  {NOTE_Fs, 4, DUR_EIGTH,},
  
  {NOTE_G,  4, DUR_EIGTH,},
  {NOTE_G,  4, DUR_EIGTH,},
  {NOTE_G,  4, DUR_SIXTEENTH,},
  {NOTE_Fs, 4, DUR_SIXTEENTH,},
  {NOTE_E,  4, DUR_SIXTEENTH,},
  {NOTE_Fs, 4, DUR_SIXTEENTH,},
  {NOTE_G,  4, DUR_EIGTH,},
  {NOTE_G,  4, DUR_EIGTH,},
  {NOTE_B,  4, DUR_SIXTEENTH,},
  {NOTE_A,  4, DUR_SIXTEENTH,},
  {NOTE_G,  4, DUR_SIXTEENTH,},
  {NOTE_A,  4, DUR_SIXTEENTH,},
  
  {NOTE_B,  4, DUR_EIGTH,},
  {NOTE_B,  4, DUR_EIGTH,},
  {NOTE_D,  5, DUR_SIXTEENTH,},
  {NOTE_C,  5, DUR_SIXTEENTH,},
  {NOTE_B,  4, DUR_SIXTEENTH,},
  {NOTE_C,  5, DUR_SIXTEENTH,},
  {NOTE_D,  5, DUR_HALF,},

}; 

struct tune tune_kleine_nachtmusik = 
{
  150,
  &tones_kleine_nachtmusik[0],
  sizeof(tones_kleine_nachtmusik) / sizeof(tones_kleine_nachtmusik[0])
};
// clang-format on

static struct {
  bool is_playing;
  int offset;
  int counter;
  struct tune *music;
} music_play;

static void pwm_interrupt(uint32_t arg)
{
  OS_API_ISetEventFlg(HOST_APL_EVENT_FLG_A, PWM_INTERRUPT_BIT);
}

/* returns a frequency value for a given note. use the NOTE_*-enums and octave numbers in the range from -5 to 5.
input is note name (0 = NOTE_C to 11 = NOTE_B) and octave number, -5 <= o <= 5, output is frequency of the desired note */
#define NOTE(i, o) notes[(i)] >> (5 - (o))

static void play_next_note(void)
{
  uint16_t freq;
  RET_API ret;

  music_play.offset++;

  if (music_play.offset == music_play.music->length) {
    music_play.is_playing = false;
    music_stopped_callback();
    dprintf("no more notes\n");
    return;
  }

  music_play.is_playing = true;

  struct tone *m = &music_play.music->tones[music_play.offset];
  music_play.counter = m->duration * 8;

  uint16_t cycle = 0;
  uint16_t duty = 0;

  if (m->note != PAUSE) {
    freq = NOTE(m->note, m->octave);
    cycle = 32768 / (freq / 10);
    /* 50% duty cycle */
    duty = cycle / 2;
    //    warning("note %d, octave %d, duration %d, frequency %d -> cycle %d\n", m->note, m->octave, m->duration, freq, cycle);
  }

  ret = Driver_PWM.Config(PWM_CH3, cycle, duty, 0xfffff);
  if (ret) {
    warning("Driver_PWM.Config() failed\n");
  }

  ret = Driver_PWM.StartPWM(PWM_CH3, PWM_1S_TIMER_ENABLE);
  if (ret) {
    warning("Driver_PWM.StartPWM() failed\n");
  }

  dprintf("freq %d\n", freq);
}

void pwm_handler(void)
{
  RET_API ret;

  music_play.counter--;
  if (music_play.counter == 1) {
    ret = Driver_PWM.Config(PWM_CH3, 0, 0, 0xfffff);
    if (ret) {
      warning("Driver_PWM.Config() failed\n");
    }
    return;
  }
  if (music_play.counter > 0) {
    return;
  }

  ret = Driver_PWM.StopPWM(PWM_CH3);
  if (ret != RET_API_NORMAL) {
    warning("Driver_PWM.Stop() failed, ret %d\n", ret);
    return;
  }

  play_next_note();
  dprintf("\n");
}

int music_start(void)
{
  int ret;

  if (music_play.is_playing == true) {
    dprintf("music still playing\n");
    return 0;
  }

  music_play.music = &tune_fuer_elise;
  music_play.offset = -1;

  if (music_tune == music_tune_kleine_nachtmusik) {
    music_play.music = &tune_kleine_nachtmusik;
  } else if (music_tune == music_tune_fuer_elise) {
    music_play.music = &tune_fuer_elise;
  } else if (music_tune == music_tune_ode_an_die_freude) {
    music_play.music = &tune_ode_an_die_freude;
  }

  ret = Driver_PWM.SetRhythmCounter(PWM_CH3, ((1637 * MS_OF_QUARTER_NOTE_FROM_BPM(music_play.music->bpm)) / 1000) / 8);
  if (ret) {
    warning("Driver_PWM.SetRhythmCounter() failed\n");
    return -1;
  }

  /* start */
  play_next_note();

  music_started_callback();

  dprintf("\n");
  return 0;
}

int music_init(void)
{
  RET_API res;

  /* set up buzzer handling */
  res = Driver_PWM.Initialize(PWM_CH3, PWM_CLOCK_32KHZ, PWM_POL_REVERSE_OFF, 0, pwm_interrupt);
  if (res) {
    warning("Driver_PWM.Initialize() failed\n");
    return -1;
  }

  dprintf("\n");
  return 0;
}
